/* 
 *    fame.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */
#include <string>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <getopt.h>

#include <module.h>
#include <video/video.h>
#include <system/system.h>
#include <input/input.h>
#include <output/output.h>
#include <frame.h>

struct Global
{
  char **env;

  string input_args;
  string video_args;
  string output_args;

  VideoModule * video_module;
  InputModule * input_module;
  OutputModule * output_module;

  Global();
  bool ParseArgs(int argc, char *argv[], char *environement[]);
  void Usage(int argc, char *argv[], char *environement[]);
};

static Global fame;

/* Options */
static const struct option longopts[]={
    {"video",1,0,'v'},
    {"input",1,0,'i'},
    {"output",1,0,'o'},
    {"help",0,0,'h'},
    {0,0,0,0}
  };
static const char * shortopts = "v:i:o:h";

bool init();
bool cleanup();
void video_callback(VideoFrame * frame);
void output_callback(void * data, int size);
void break_handler(int signal);

Global::Global()
{
  input_args = "";
  video_args = "";
  output_args = "";

  video_module = 0;
  input_module = 0;
  output_module = 0;
}

bool Global::ParseArgs(int argc, char * argv[], char *environement[])
{
  int c = 0;
  int index;

  env = environement;

  while((c = getopt_long(argc,argv,shortopts,longopts,&index)) != -1)
  {
    switch(c) 
    {
        case 'v':
	  video_args = optarg;
  	break;
        case 'i':
          input_args = optarg;
  	break;
        case 'o':
          output_args = optarg;
  	break;
	case 'h':
	return(false);
        default: 
        return(false);
    }
  }  
  return(true);
}

void Global::Usage(int argc, char *argv[], char *env[])
{
  cerr << "usage : " << argv[0] << " [options]" << endl;
  cerr << "   options can be" << endl;
  cerr << "      --input  or -i args              : arguments to give to the input module (-i help gives info on all input modules)" << endl;
  cerr << "      --video  or -v args              : arguments to give to the video module (-v help gives info on all video modules)" << endl;
  cerr << "      --output or -o args              : arguments to give to the output module (-o help gives info on all output modules)" << endl;
  cerr << "      --help   or -h                   : print this help" << endl;
}

bool init()
{
  ModulePath * module_path;
  string full_path;
  ModuleList * module_list;
  char const * name;

  module_path = new ModulePath(fame.env);
  cerr << "module path is " << module_path->Path() << endl;

  /* Load the input module */
  cerr << "input module: ";

  full_path = string(module_path->Path()) + string("/input");
  module_list = new ModuleList(full_path.c_str());

  while((name = module_list->Next()))
  {
    fame.input_module = new InputModule(name, fame.input_args.c_str());
#ifdef DEBUG_MODULE
    cerr << "init: Trying " << fame.input_module->identify() << " input module " << ((fame.input_module->status())?" : success":" : failure") << endl;
#endif
    if(fame.input_module->status())
      break;

    delete fame.input_module;

    fame.input_module = 0;
  }

  delete module_list;

  cerr << ((fame.input_module)?fame.input_module->identify():"not found") << endl;

  /* Load the video module */
  cerr << "video module: ";

  full_path = string(module_path->Path()) + string("/video");
  module_list = new ModuleList(full_path.c_str());

  while((name = module_list->Next()))
  {
    fame.video_module = new VideoModule(name, fame.video_args.c_str());
#ifdef DEBUG_MODULE
    cerr << "init: Trying " << fame.video_module->identify() << " video module " << ((fame.video_module->status())?" : success":" : failure") << endl;
#endif
    if(fame.video_module->status())
      break;

    delete fame.video_module;

    fame.video_module = 0;
  }

  delete module_list;

  cerr << ((fame.video_module)?fame.video_module->identify():"not found") << endl;

  /* Load the output module */
  cerr << "output module: ";

  full_path = string(module_path->Path()) + string("/output");
  module_list = new ModuleList(full_path.c_str());

  while((name = module_list->Next()))
  {
    fame.output_module = new OutputModule(name, fame.output_args.c_str());
#ifdef DEBUG_MODULE
    cerr << "init: Trying " << fame.output_module->identify() << " output module " << ((fame.output_module->status())?" : success":" : failure") << endl;
#endif
    if(fame.output_module->status())
      break;

    delete fame.output_module;

    fame.output_module = 0;
  }

  delete module_list;

  cerr << ((fame.output_module)?fame.output_module->identify():"not found") << endl;

  delete module_path;

  /* Test if module initialization succeeded */
  if(!fame.input_module || 
     !fame.video_module ||
     !fame.output_module)
    return(false);

  /* Plug the modules together */
  fame.video_module->SetCallback(output_callback);
  fame.input_module->SetCallback(video_callback);

  return(true);
}

bool cleanup()
{
  /* Unhook signals */
  signal(SIGINT, SIG_DFL);

  if(fame.input_module) delete fame.input_module;
  if(fame.video_module) delete fame.video_module;
  if(fame.output_module) delete fame.output_module;

  return(true);
}

void break_handler(int signal)
{
  /* Handle the break (Ctrl-C) signal */
  if(signal != SIGINT) return;

  cleanup();

  cerr << "break signal caught" << endl;

  exit(0);
}

void video_callback(VideoFrame * frame)
{
  fame.video_module->EncodeFrame(frame);
}

void output_callback(void * data, int size)
{
  fame.output_module->Write(data, size);
}

int main(int argc, char * argv[], char *env[])
{
    /* Hook to the break signal */
    signal(SIGINT, break_handler);

    if(!fame.ParseArgs(argc, argv, env))
    {
        fame.Usage(argc, argv, env);
        exit(0);
    }

    if(!init())
    {
        cerr << "initialization failed" << endl;
        fame.Usage(argc, argv, env);
	cleanup();
	exit(-1);
    }

    /* Enter an endless loop */
    while(true)
      sleep(1);
    
    cleanup();

    return(0);
}


