/* 
 *    module.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <dlfcn.h>
#include <stdio.h>
#include <unistd.h>

#include <module.h>

Module::Module(char const * filename, char const * args)
{
  int argc;
  char ** argv;

  good = false;

  /* Get a handle to the module */
  module = dlopen(filename, RTLD_NOW);

  if(!module) return;

  /* Load the identifier */
  name = (char *) get_symbol("module_id");

  if(!name) return;

  /* Separate the args string into C-like argc, argv[] argument list*/
  char *str, *a;

  argc = 0;
  argv = 0;

  a = new char[strlen(args) + 1];
  strcpy(a, args);
  str = strtok(a, ",");

  if(str)
  {
    argc++;
    argv = (char **) malloc(sizeof(char **));
    argv[0] = new char [strlen(str) + 1];
    strcpy(argv[0], str);

    while((str = strtok(0, ",")))
    {
      argc++;
      argv = (char **) realloc(argv, argc * sizeof(char **));
      argv[argc - 1] = new char [strlen(str) + 1];
      strcpy(argv[argc - 1], str);
    }
  }
  delete a;

  /* Load init function */
  module_init_f init;

  init = (module_init_f) get_symbol("module_init");
  
  if(!init) return;

  context = init(argc, argv);

  if(!context) return;

  /* Release the args */
  for(register int i = 0; i < argc; i++)
    delete argv[i];
  free(argv);

  good = true;
}

Module::~Module()
{
  if(module)
  {
    module_close_f close;

    close = (module_close_f) get_symbol("module_close");
    if(close) close(context);

    // TEMP
    // BUG
    //    dlclose(module);
  }
}

char const * Module::identify()
{
  return(name);
}

bool Module::status()
{
  return(good);
}

void * Module::get_symbol(char const * symbol)
{
  return(dlsym(module, symbol));
}

ModulePath::ModulePath(char * env[])
{
  path = 0;

#ifdef MODULE_PATH
  path = (char *) malloc(strlen(MODULE_PATH) + 1);
  strcpy(path, MODULE_PATH);
#endif

  if(env)
  {
    /* Parse the environnement for the FAME_MODULE_PATH variable */
    for(register int i = 0; env[i]; i++)
      if(!strncmp(env[i], "FAME_MODULE_PATH=", strlen("FAME_MODULE_PATH=")))
      {
	if(path) free(path);
	path = (char *) malloc(strlen(env[i]) - strlen("FAME_MODULE_PATH=") + 1);
	strcpy(path, env[i] + strlen("FAME_MODULE_PATH="));
      }
  }
}

char const * ModulePath::Path() const
{
  return(path);
}

ModulePath::~ModulePath()
{
  free(path);
}

ModuleList::ModuleList(char const * path)
{
  struct dirent * entry;
  DIR * dir;

  module_list = (char **) malloc(sizeof(char *));
  module_list[0] = 0;

  dir = opendir(path);
  if(dir == NULL) return;

  count = 0;
  while((entry = readdir(dir)) != NULL)
  {
    char * str;
    void * mod;

    str = new char [strlen(path) + 1 + strlen(entry->d_name) + 1];
    strcpy(str, path);
    strcat(str, "/");
    strcat(str, entry->d_name);
    mod = dlopen(str, RTLD_LAZY);
    
    if(mod)
    {
      if(dlsym(mod, "module_id"))
      {
	count++;
	module_list = (char **) realloc(module_list, (count+1) * sizeof(char *));
	module_list[count - 1] = str;
	module_list[count] = 0;
      }
      else
      {
#ifdef DEBUG_MODULE
	fprintf(stderr, "ModuleList: dlsym %s\n", dlerror());
#endif    
	delete str;
      }

      dlclose(mod);
    }
    else
    {
#ifdef DEBUG_MODULE
      fprintf(stderr, "ModuleList: dlopen %s\n", dlerror());
#endif    
      delete str;
    }
  }

  closedir(dir);

  count = 0;
}

char const * ModuleList::Next()
{
  if(module_list[count])
    return(module_list[count++]);
  else
    return(0);
}

ModuleList::~ModuleList()
{
  int i;

  for(i = 0; module_list[i]; i++)
  {
    delete module_list[i];
  }
  delete module_list;
}
