/* 
 *    mpegnc.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <iostream>
#include <string>
#include <fstream>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sockstream.h>
#include <sockinet.h>
#include <output/sockmcast.h>
#include <output/rtp.h>
#include <pthread.h>

#define USE_RTP

/* Buffer for incomming packets */
#define DEFAULT_BUFFER_SIZE 2 * 1024

#define SLICE_MIN_START_CODE 0x00000101
#define SLICE_MAX_START_CODE 0x000001af
#define SEQ_END1_CODE        0x000001b7
#define SEQ_END2_CODE        0x000001b9
#define SEQ_START_CODE       0x000001b3
#define PICT_START_CODE      0x00000100

enum mpeg_type { mpeg_autodetect, mpeg_video, mpeg_audio, mpeg_system };

struct Args {
  char * address;
  unsigned long int port;

  unsigned long int buffer_size;
  bool verbose;
  mpeg_type type;
  
  Args(int argc, char ** argv);

private:
  void usage(char const * name);
};

struct output_thread_args {
  ostream * os;
  unsigned long int buffer_size;
  bool verbose;
  mpeg_type type;
};

struct input_thread_args {
  istream * is;
  unsigned int buffer_size;
  bool verbose;
  mpeg_type type;
};

bool timer_activated;
struct itimerval t;
unsigned long int ucountdown;

// Audio header
const int audio_frequencies[2][3]=
{
  {44100,48000,32000}, // MPEG 1
  {22050,24000,16000}  // MPEG 2
};

const int audio_bitrate[2][3][15]=
{
  // MPEG 1
  {{0,32,64,96,128,160,192,224,256,288,320,352,384,416,448},
   {0,32,48,56,64,80,96,112,128,160,192,224,256,320,384},
   {0,32,40,48,56,64,80,96,112,128,160,192,224,256,320}},

  // MPEG 2
  {{0,32,48,56,64,80,96,112,128,144,160,176,192,224,256},
   {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160},
   {0,8,16,24,32,40,48,56,64,80,96,112,128,144,160}}

};

const float video_framerate[8] =
{
  23.976, 24, 25, 29.97, 30, 50, 59.94, 60
};

// Argument parser
Args::Args(int argc, char ** argv)
{
  register int i;
  
  address = 0;
  port = 0;
  verbose = false;
  buffer_size = DEFAULT_BUFFER_SIZE;
  type = mpeg_autodetect;
  
  if(argc < 2) usage(argv[0]);
  
  for(i = 1; i < argc && argv[i][0] == '-'; i++)
  {
    switch(argv[i][1])
    {
      // Verbose mode
      case 'v':
	verbose = true;
      break;
	
      // Buffer size
      case 'b':
	i++;
	if(i >= argc) usage(argv[0]);
	buffer_size = atoi(argv[i]);
      break;

      // Help
      case 'h':
	usage(argv[0]);
      break;

      // Type
      case 't':
	i++;
	if(string(argv[i]) == "audio") type = mpeg_audio;
	if(string(argv[i]) == "video") type = mpeg_video;
      break;
	
      default:
	usage(argv[0]);
      break;
    }
  }

  if(!strchr(argv[i], ':')) usage(argv[0]);
  
  address = argv[i];	
  port = atoi(strchr(argv[i], ':') + sizeof(char));
  *strchr(argv[i], ':') = '\0';
  address = argv[i];
}

void Args::usage(char const * name)
{
  cerr << "usage : " << name << " [-v] [-b size] [-h] [-l] address:port" << endl;
  exit(0);
}

inline bool audio_header(unsigned long code)
{
  return((code & 0xfff00000) == 0xfff00000 &&  // Sync bits
	 (code & 0x0000f000) != 0x00000000 &&  // Bitrate not 0
	 (code & 0x0000f000) != 0x0000f000 &&  // Bitrate not 15
	 (code & 0x00000c00) != 0x00000c00 &&  // Frequency not 3
	 (code & 0x00060000) != 0x00000000);   // Layer not 4
}

void mpeg_connect(istream & is, ostream & os, unsigned long buffer_size = DEFAULT_BUFFER_SIZE, mpeg_type type = mpeg_autodetect, bool verbose = false, bool use_rtp_input = false, bool use_rtp_output = false)
{
  register unsigned long int code;
  register unsigned char * i_ptr;
  unsigned char * i_stop;
  unsigned char * i_base = new unsigned char[buffer_size];
  register unsigned char * o_ptr;
  unsigned char * o_stop;
  unsigned char * o_base = new unsigned char[buffer_size];

  struct timeval tv1;
  struct timeval tv2;
  struct timezone tz;

  register unsigned i;
  unsigned totaltime, totalframes;
  int usec, frameutime;

  // A macro to get the next code and check for buffer bounds
#define NEXT_CODE()                               \
{                                                 \
    if(i_ptr == i_stop)                           \
    {                                             \
      is.read(i_base, buffer_size);               \
      i_ptr = i_base;                             \
    }                                             \
                                                  \
    if(o_ptr == o_stop)                           \
    {                                             \
      os.write(o_base, buffer_size);              \
      o_ptr = o_base;                             \
    }                                             \
                                                  \
    *o_ptr++ = (unsigned char) (code >> 24);      \
    code <<= 8;                                   \
    code |= *i_ptr++;                             \
}

  // Initialize to an invalid code
  code = 0x55555555;

  i_ptr = i_base;
  i_stop = i_base + buffer_size;
  o_ptr = o_base;
  o_stop = o_base + buffer_size;

  // Seek start code
  while(true)
  {
    if(code == SEQ_START_CODE && (type == mpeg_autodetect || type == mpeg_video))
    {
      type = mpeg_video;
      break;
    }

    if(audio_header(code) &&
       (type == mpeg_autodetect || type == mpeg_audio))
    {
      type = mpeg_audio;
      break;
    }

    if(i_ptr == i_stop)
    {
      is.read(i_base, buffer_size);
      i_ptr = i_base;
    }

    code <<= 8;
    code |= *i_ptr++;
  }


  if(type == mpeg_audio)
  {
    unsigned layer, version, frequency, bitrate, mode, padding, framesize;

    gettimeofday(&tv1, &tz);
    usec = 0;
    totaltime = 0;
    totalframes = 0;

    while(is.good() && os.good())
    {
      layer = 4 - ((code >> 17) & 3);
      version = ((code>> 19) & 1) ^ 1;
      padding = (code >> 9) & 1;
      frequency = audio_frequencies[version][((code >> 10) & 3)];
      bitrate = audio_bitrate[version][layer-1][((code >> 12) & 15)];
      mode = (code >> 6) & 3;

      // framesize
      if(layer==1)
      {
	framesize = 12000 * bitrate / frequency;
	if(frequency==0 && padding) framesize++;
	framesize <<= 2;
      }
      else
      {
	framesize = 144000 * bitrate / (frequency<<version);
	if(padding) framesize++;
      }

      frameutime = 8000 * framesize / bitrate;

      usec += frameutime;

      if(verbose)
      {
	cerr << "frame time = " << frameutime << endl;
	cerr << "MPEG " << dec << (1 + version) << " ";
	cerr << "layer " << layer << " ";
	cerr << frequency << "Hz ";
	cerr << bitrate << "kbits/s ";
	cerr << "mode = " << mode << " ";
	cerr << "size = " << framesize << endl;
      }

      // copy frame
      for(i = 0; i < framesize; i++)
	NEXT_CODE();

      while(!audio_header(code))
	NEXT_CODE();

      gettimeofday(&tv2, &tz);

      usec -= 1000000 * (tv2.tv_sec - tv1.tv_sec) + (tv2.tv_usec - tv1.tv_usec);
      tv1 = tv2;

      if(verbose)
	cerr << "frame " << totalframes << " time " << totaltime / 1000000.0 << endl;

      // time sync
      if(usec > 0)
	usleep(usec);

      os.write(o_base, o_ptr - o_base);
      o_ptr = o_base;
      os.flush();
      totalframes++;
      totaltime += frameutime;

      gettimeofday(&tv2, &tz);
      usec -= 1000000 * (tv2.tv_sec - tv1.tv_sec) + (tv2.tv_usec - tv1.tv_usec);
      tv1 = tv2;
    }
  }

  if(type == mpeg_video)
  {
    if(verbose)
      cerr << "SEQUENCE HEADER - video only stream" << endl;

    gettimeofday(&tv1, &tz);
    frameutime = 0;
    usec = 0;
    totaltime = 0;
    totalframes = 0;

    while(is.good() && os.good())
    {
      if(code == SEQ_START_CODE)                                       // Sequence header
      {
	unsigned header_size = 8;

	for(i = 0; i < header_size; i++)
	{
	  NEXT_CODE();

	  switch(i)
	  {
	    case 3:
	      frameutime = (unsigned) (1000000.0 / video_framerate[code & 0xf]); // frame rate
	    break;

	    case 7:
	      if(code & 2) header_size += 64; // intra matrix
	      if(code & 1) header_size += 64; // non-intra matrix
	    break;
	  }
	}
      }
      else
      if(code == PICT_START_CODE)
      {
	for(i = 0; i < 4; i++)
	  NEXT_CODE();

	usec += frameutime;

	gettimeofday(&tv2, &tz);

	usec -= 1000000 * (tv2.tv_sec - tv1.tv_sec) + (tv2.tv_usec - tv1.tv_usec);
	tv1 = tv2;

	if(verbose)
	  cerr << "frame " << totalframes << " time " << totaltime / 1000000.0 << endl;

	// time sync
	if(usec > 0)
	  usleep(usec);

	totalframes++;
	totaltime += frameutime;

	gettimeofday(&tv2, &tz);
	usec -= 1000000 * (tv2.tv_sec - tv1.tv_sec) + (tv2.tv_usec - tv1.tv_usec);
	tv1 = tv2;
      }
      if(code >= SLICE_MIN_START_CODE && code <= SLICE_MAX_START_CODE) // Slice header
      {
	for(i = 0; ; i++)
	{
	  NEXT_CODE();

	  if((code & 0xffffff00) == 0x00000100) // found a header
	    break;
	}
	
	os.write(o_base, o_ptr - o_base);
	o_ptr = o_base;
	os.flush();
      }
      else                                                            // Unknown data
      {
	NEXT_CODE();
      }
    }
  }

  delete i_base;
  delete o_base;
}

void * input_thread(void * data)
{
  input_thread_args * it_args = (input_thread_args *) data;

  mpeg_connect(*it_args->is, cout, it_args->buffer_size, it_args->type, it_args->verbose, false, false);

  return(0);
}

void * output_thread(void * data)
{
  output_thread_args * ot_args = (output_thread_args *) data;
  
  mpeg_connect(cin, *ot_args->os, ot_args->buffer_size, ot_args->type, ot_args->verbose, false, false);

  return(0);
}

void break_handler(int signal)
{
  /* Handle the break (Ctrl-C) signal */

  cerr << "break signal caught" << endl;

  if(signal != SIGINT) return;

  exit(0);
}


int main (int argc, char **argv)
{
  pthread_t it, ot;
  Args args(argc, argv);
  isockmcast is;
  osockmcast os;

  input_thread_args it_args;
  output_thread_args ot_args;

  is->reuseaddr(true);
  is->linebuffered(0);
  is->bind(args.address, args.port);

  os->reuseaddr(true);
  os->linebuffered(0);
  os->bind(args.address, args.port);
  os->connect(args.address, args.port);

  /* Hook to the break signal */
  signal(SIGINT, break_handler);

  it_args.is = &is;
  it_args.buffer_size = args.buffer_size;
  it_args.verbose = args.verbose;
  it_args.type = args.type;
  pthread_create(&it, NULL, input_thread, (void *) &it_args);
  ot_args.os = &os;
  ot_args.buffer_size = args.buffer_size;
  ot_args.verbose = args.verbose;
  ot_args.type = args.type;
  pthread_create(&ot, NULL, output_thread, (void *) &ot_args);

  while(true)
    usleep(1000);
}
