/* 
 *    record.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <iostream>
#include <string>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/ioctl.h>
#include <linux/soundcard.h>

#define DEFAULT_BUFFER_SIZE 2 * 1024
#define DEFAULT_RESOLUTION 16
#define DEFAULT_SAMPLE_RATE 44100
#define DEFAULT_CHANNELS 2

struct waveheader_t
{
  unsigned char  RIFF[4];          // "RIFF"
  unsigned long  size;             // Size of data to follow
  unsigned char  WAVE[4];          // "WAVE"
  unsigned char  fmt_[4];          // "fmt "
  unsigned long  _16;              // 16
  unsigned short _1;               // 1
  unsigned short channels;         // Number of Channels
  unsigned long  sample_rate;      // Sample Rate
  unsigned long  bps;              // Bytes / sec
  unsigned short block_align;      // 1
  unsigned short bits;             // Bits / sample
  unsigned char  data[4];          // "data"
  unsigned long  data_size;        // Number of Samples
  
  waveheader_t()
  {
    RIFF[0] = 'R';
    RIFF[1] = 'I';
    RIFF[2] = 'F';
    RIFF[3] = 'F';
    WAVE[0] = 'W';
    WAVE[1] = 'A';
    WAVE[2] = 'V';
    WAVE[3] = 'E';
    fmt_[0] = 'f';
    fmt_[1] = 'm';
    fmt_[2] = 't';
    fmt_[3] = ' ';
    data[0] = 'd';
    data[1] = 'a';
    data[2] = 't';
    data[3] = 'a';
    _16 = 16;
    _1 = 1;
    size = 0;
    channels = 0;
    sample_rate = 0;
    bps = 0;
    block_align = 0;
    bits = 0;
    data_size = 0;
  }

  void set_channels(int n)
  {
    channels = n;
    block_align = channels * bits / 8;
    bps = sample_rate * channels * bits / 8;
  }

  void set_sample_rate(int n)
  {
    sample_rate = n;
    block_align = channels * bits / 8;
    bps = sample_rate * channels * bits / 8;
  }

  void set_bits(int n)
  {
    bits = n;
    block_align = channels * bits / 8;
    bps = sample_rate * channels * bits / 8;
  }

  void set_size(int n)
  {
    data_size = n;
    size = data_size + sizeof(waveheader_t) - 8;
  }
};

struct Args {
  Args(int argc, char ** argv);

  int bits;
  int sample_rate;
  int channels;
  int buffer_size;

private:

  void usage(char const * name);
};

Args::Args(int argc, char ** argv)
{
  register int i;

  buffer_size = DEFAULT_BUFFER_SIZE;
  bits = DEFAULT_RESOLUTION;
  sample_rate = DEFAULT_SAMPLE_RATE;
  channels = DEFAULT_CHANNELS;

  for(i = 1; i < argc && argv[i][0] == '-'; i++)
  {
    switch(argv[i][1])
    {
      // Help
      case 'h':
	usage(argv[0]);
      break;

      // Buffer size
      case 'b':
	i++;
	if(i >= argc) usage(argv[0]);
	buffer_size = atoi(argv[i]);
      break;

      // Resolution (bits)
      case 'r':
	i++;
	if(i >= argc) usage(argv[0]);
	bits = atoi(argv[i]);
      break;

      // Sample rate
      case 's':
	i++;
	if(i >= argc) usage(argv[0]);
	sample_rate = atoi(argv[i]);
      break;

      // Type
      case 't':
	i++;
	if(string(argv[i]) == "mono") channels = 1;
	if(string(argv[i]) == "stereo") channels = 2;
      break;

      default:
	usage(argv[0]);
      break;
    }
  }
}

void Args::usage(char const * name)
{
  cerr << "usage : " << name << "[-h] [-b size] [-r bits] [-s sample_rate] [-t mono | stereo]" << endl;
  exit(0);
}

void break_handler(int signal)
{
  /* Handle the break (Ctrl-C) signal */

  cerr << "break signal caught" << endl;

  if(signal != SIGINT) return;

  exit(0);
}


int main (int argc, char **argv)
{
  Args args(argc, argv);

  /* Hook to the break signal */
  signal(SIGINT, break_handler);

  int audio_fd;
  int format;
  int stereo;
  int frequency;

  audio_fd = open("/dev/dsp", O_RDONLY);

  ioctl(audio_fd, SNDCTL_DSP_GETFMTS, &format);

  cerr << "Supported formats" << endl;
  cerr << "=================" << endl;

  if(format & AFMT_U8) cerr << "Unsigned 8" << endl;
  if(format & AFMT_S8) cerr << "Signed 8" << endl;
  if(format & AFMT_U16_LE) cerr << "Unsigned 16 Low-Endian" << endl;
  if(format & AFMT_U16_BE) cerr << "Unsigned 16 Big-Endian" << endl;
  if(format & AFMT_S16_LE) cerr << "Signed 16 Low-Endian" << endl;
  if(format & AFMT_S16_BE) cerr << "Signed 16 Big-Endian" << endl;
  
  format = (args.bits == 8)?AFMT_U8:AFMT_S16_LE;
  if(ioctl(audio_fd, SNDCTL_DSP_SETFMT, &format) < 0) return 0; 

  stereo = args.channels - 1;
  ioctl(audio_fd, SNDCTL_DSP_STEREO, &stereo);

  frequency = args.sample_rate;
  ioctl(audio_fd, SOUND_PCM_WRITE_RATE, &frequency);

  char * buffer = new char [args.buffer_size];
  waveheader_t wh;

  wh.set_sample_rate(args.sample_rate);
  wh.set_channels(args.channels);
  wh.set_bits(args.bits);
  wh.set_size(args.buffer_size);
  write(1, &wh, sizeof(waveheader_t));

  while(true)
  {
    read(audio_fd, buffer, args.buffer_size);
    write(1, buffer, args.buffer_size);
  }

  close(audio_fd);
}
