/* 
 *    catstream.cpp
 *
 *	Copyright (C) Vivien Chappelier - 2000
 *
 *  This file is part of fame, a free MPEG encoder.
 *	
 *  fame is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  fame is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include <iostream>
#include <string>
#include <fstream>
/*
#include <network/sstream.h>
*/
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <sockstream.h>
#include <sockinet.h>
#include <pthread.h>

/* Timeout before connection is considered lost */
#define TIMEOUT 100000

/* Buffer for incomming packets */
/* WARNING: This buffer may overflow */
#define DEFAULT_BUFFER_SIZE 256 * 1024

struct Args {
  char * address;
  unsigned long int port;


  unsigned long int buffer_size;
  bool verbose;
  bool listen;
  
  Args(int argc, char ** argv);

private:
  void usage(char const * name);
};

isockinet is(sockbuf::sock_dgram);
osockinet os(sockbuf::sock_dgram);
bool timer_activated;
struct itimerval t;
unsigned long int ucountdown;


// Argument parser
Args::Args(int argc, char ** argv)
{
  register int i;
  
  address = 0;
  port = 0;
  verbose = false;
  listen = false;
  buffer_size = DEFAULT_BUFFER_SIZE;
  
  if(argc < 2) usage(argv[0]);
  
  for(i = 1; i < argc && argv[i][0] == '-'; i++)
  {
    switch(argv[i][1])
    {
      // Verbose mode
      case 'v':
	verbose = true;
      break;
	
      // Buffer size
      case 'b':
	i++;
	if(i >= argc) usage(argv[0]);
	buffer_size = atoi(argv[i]);
      break;

      // Help
      case 'h':
	usage(argv[0]);
      break;

      // Listen
      case 'l':
	listen = true;
      break;
	
      default:
	usage(argv[0]);
      break;
    }
  }

  if(!strchr(argv[i], ':')) usage(argv[0]);
  
  address = argv[i];	
  port = atoi(strchr(argv[i], ':') + sizeof(char));
  *strchr(argv[i], ':') = '\0';
  address = argv[i];
}

void Args::usage(char const * name)
{
  cerr << "usage : " << name << " [-v] [-b size] [-h] [-l] address:port" << endl;
  exit(0);
}

/* Check for a valid mpeg video header */ 
bool video_header(unsigned char * buffer)
{
  return(buffer[0] == 0x00 && 
	 buffer[1] == 0x00 && 
	 buffer[2] == 0x01 && 
	 buffer[3] == 0xb3);
}

/* Check for a valid mpeg audio header */ 
bool audio_header(unsigned char * buffer)
{
  return(buffer[0] == 0xff && 
	 ((buffer[1] & 0xf0) == 0xf0) &&
	 ((buffer[2] & 0xf0) != 0xf0) && 
	 ((buffer[2] & 0xf0) != 0x00));
}

long seek_next_video_frame(unsigned char * buffer, unsigned long int count)
{
  register unsigned long i;

  for(i = 0; i < count; i++)
    if(buffer[i] == 0x00 && 
       buffer[i+1] == 0x00 && 
       buffer[i+2] == 0x01)
      break;

  return(i);
}

void * input_thread(void * data)
{
  Args * args = (Args *) data;
  bool sequence_started;
  unsigned char * buffer;

  buffer = new unsigned char[args->buffer_size / 2];

  if(args->verbose)
    cerr << "Using input buffer of " << args->buffer_size / 2 << " bytes" << endl;

  sequence_started = false;

  while(true)
  {     
    /* Read from network to console */

      is.read(buffer, args->buffer_size / 2);
      is.clear();

      if(args->verbose && is.gcount() != 0 && video_header(buffer))
	cerr << "Video startcode" << endl;

      if(args->verbose && is.gcount() != 0 && audio_header(buffer))
	cerr << "Audio startcode" << endl;

      if(!sequence_started && video_header(buffer))
      {
	sequence_started = true;
	if(args->verbose)
	  cerr << "Sequence started : Video stream" << endl;
      }

      if(!sequence_started && audio_header(buffer))
      {
	sequence_started = true;
	if(args->verbose)
	  cerr << "Sequence started : Audio stream" << endl;
      }

      if(sequence_started) 
	cout.write(buffer, is.gcount());
    
      if(args->verbose && sequence_started && is.gcount() != 0)
	cerr << "Wrote " << is.gcount() << " bytes" << endl;

      cout.flush();
      usleep(1000);
  }

  delete buffer;
  
  return(0);
}

void * output_thread(void * data)
{
  Args * args = (Args *) data;
  unsigned char * buffer;
  unsigned long pos;

  buffer = new unsigned char[args->buffer_size / 2];

  if(args->verbose)
    cerr << "Using output buffer of " << args->buffer_size / 2 << " bytes" << endl;

  cin.read(buffer, args->buffer_size / 2);
  cin.clear();

  do
  {
    pos = seek_next_video_frame(buffer, args->buffer_size/2);
    memmove(buffer, buffer+pos, args->buffer_size/2 - pos);
    cin.read(buffer + args->buffer_size/2 - pos, pos);
    cin.clear();
  }
  while(!video_header(buffer));

  switch(buffer[7] & 0x0f)
  {
    case 1:
      ucountdown = (unsigned long int) (1000000. / 23.97);
    break;
    case 2:
      ucountdown = (unsigned long int) (1000000. / 24);
    break;
    case 3:
      ucountdown = (unsigned long int) (1000000. / 25);
    break;
    case 4:
      ucountdown = (unsigned long int) (1000000. / 29.9);
    break;
    case 5:
      ucountdown = (unsigned long int) (1000000. / 30);
    break;
    default:
      cerr << "unknown frame rate" << endl;
    return(0);
  }

  if(args->verbose)
    cerr << "Frame rate " << 1000000.0 / ucountdown << endl;

  memmove(buffer, buffer+pos, args->buffer_size/2 - pos);
  cin.read(buffer + args->buffer_size/2 - pos, pos);
  cin.clear();

  /* Initialize the timer */
  timer_activated = false;
  t.it_value.tv_sec = 0;
  t.it_value.tv_usec = ucountdown;
  setitimer(ITIMER_REAL, &t, NULL);

  while(buffer[3] != 0xb7 && buffer[3] != 0xb9)
  {     
    /* Write to network from console */
      // Skip possible current sequence head code, GOP, frame and slice header
      pos = seek_next_video_frame(buffer+36, args->buffer_size/2-36) + 36;

      while(!timer_activated) wait(0);
      timer_activated = false;
      os.write(buffer, pos);
      os.flush();
      memmove(buffer, buffer+pos, args->buffer_size/2 - pos);
      cin.read(buffer + args->buffer_size/2 - pos, pos);
  }

  delete buffer;
  
  return(0);
}

void timer_handler(int signal)
{
  /* Handle the timer signal */

  if(signal != SIGALRM) return;

  timer_activated = true;
  t.it_value.tv_sec = 0;
  t.it_value.tv_usec = ucountdown;
  setitimer(ITIMER_REAL, &t, NULL);
}

void break_handler(int signal)
{
  /* Handle the break (Ctrl-C) signal */

  cerr << "break signal caught" << endl;

  if(signal != SIGINT) return;

  exit(0);
}

int main (int argc, char **argv)
{
  char * strbufin, *strbufout;
  pthread_t it, ot;
  Args args(argc, argv);

  sockinetaddr sa(args.address, args.port);

  is->reuseaddr(true);
  is->bind(args.address, args.port);
  is->connect(sa);
  //  is->multicast(args.address, args.port, true);
  os->reuseaddr(true);
  os->bind(args.address, args.port);
  os->connect(sa);
  //  os->multicast(args.address, args.port, true);

  cout << is->localhost() << endl;

  os << 1 << endl;
  int toto;
  is >> toto;
  cout << toto << endl;
/*
  strbufin = new char[args.buffer_size];
  strbufout = new char[args.buffer_size];
  args.is.rdbuf()->setbuf(strbufin, args.buffer_size);
  args.os.rdbuf()->setbuf(strbufout, args.buffer_size);
*/

  pthread_create(&it, NULL, input_thread, (void *) &args);
  pthread_create(&ot, NULL, output_thread, (void *) &args);

  /* Hook to the break signal */
  signal(SIGINT, break_handler);

  /* Hook to the timer signal */
  signal(SIGALRM, timer_handler);

  while(true)
    usleep(1000);

  is->multicast(args.address, args.port, false);
  os->multicast(args.address, args.port, false);

/*
  delete strbufin;
  delete strbufout;
*/
}
